/*
 * Java port of Bullet (c) 2008 Martin Dvorak <jezek2@advel.cz>
 *
 * Bullet Continuous Collision Detection and Physics Library
 * Copyright (c) 2003-2007 Erwin Coumans  http://continuousphysics.com/Bullet/
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

package javabullet.collision.shapes;

import java.util.ArrayList;
import java.util.List;
import javabullet.BulletStack;
import javabullet.linearmath.AabbUtil2;
import javabullet.linearmath.MiscUtil;
import javabullet.linearmath.VectorUtil;
import javax.vecmath.Vector3f;

/**
 * OptimizedBvh store an AABB tree that can be quickly traversed on CPU (and SPU,GPU in future).
 * 
 * @author jezek2
 */
public class OptimizedBvh {

	protected final BulletStack stack = BulletStack.get();
	
	private static final boolean DEBUG_TREE_BUILDING = false;
	private static int gStackDepth = 0;
	private static int gMaxStackDepth = 0;
	
	private static int maxIterations = 0;
	
	// Note: currently we have 16 bytes per quantized node
	public static final int MAX_SUBTREE_SIZE_IN_BYTES = 2048;

	// 10 gives the potential for 1024 parts, with at most 2^21 (2097152) (minus one
	// actually) triangles each (since the sign bit is reserved
	public static final int MAX_NUM_PARTS_IN_BITS = 10;
	
	////////////////////////////////////////////////////////////////////////////

	private final List<OptimizedBvhNode> leafNodes = new ArrayList<OptimizedBvhNode>();
	private final List<OptimizedBvhNode> contiguousNodes = new ArrayList<OptimizedBvhNode>();

	//private List<QuantizedBvhNode> quantizedLeafNodes = new ArrayList<QuantizedBvhNode>();
	//private List<QuantizedBvhNode> quantizedContiguousNodes = new ArrayList<QuantizedBvhNode>();
	
	private int curNodeIndex;

	// quantization data
	private boolean useQuantization;
	private final Vector3f bvhAabbMin = new Vector3f();
	private final Vector3f bvhAabbMax = new Vector3f();
	private final Vector3f bvhQuantization = new Vector3f();
	
	protected TraversalMode traversalMode;
	protected final List<BvhSubtreeInfo> SubtreeHeaders = new ArrayList<BvhSubtreeInfo>();
	// This is only used for serialization so we don't have to add serialization directly to btAlignedObjectArray
	protected int subtreeHeaderCount;

	// two versions, one for quantized and normal nodes. This allows code-reuse while maintaining readability (no template/macro!)
	// this might be refactored into a virtual, it is usually not calculated at run-time
	public void setInternalNodeAabbMin(int nodeIndex, Vector3f aabbMin) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//quantizeWithClamp(&m_quantizedContiguousNodes[nodeIndex].m_quantizedAabbMin[0] ,aabbMin);
		}
		else {
			contiguousNodes.get(nodeIndex).aabbMinOrg.set(aabbMin);
		}
	}

	public void setInternalNodeAabbMax(int nodeIndex, Vector3f aabbMax) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//quantizeWithClamp(&m_quantizedContiguousNodes[nodeIndex].m_quantizedAabbMax[0],aabbMax);
		}
		else {
			contiguousNodes.get(nodeIndex).aabbMaxOrg.set(aabbMax);
		}
	}
	
	public Vector3f getAabbMin(int nodeIndex) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//return unQuantize(&m_quantizedLeafNodes[nodeIndex].m_quantizedAabbMin[0]);
		}
		//non-quantized
		return leafNodes.get(nodeIndex).aabbMinOrg;

	}

	public Vector3f getAabbMax(int nodeIndex) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//return unQuantize(&m_quantizedLeafNodes[nodeIndex].m_quantizedAabbMax[0]);
		}
		//non-quantized
		return leafNodes.get(nodeIndex).aabbMaxOrg;
	}
	
	public void setInternalNodeEscapeIndex(int nodeIndex, int escapeIndex) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//m_quantizedContiguousNodes[nodeIndex].m_escapeIndexOrTriangleIndex = -escapeIndex;
		}
		else {
			contiguousNodes.get(nodeIndex).escapeIndex = escapeIndex;
		}
	}

	public void mergeInternalNodeAabb(int nodeIndex, Vector3f newAabbMin, Vector3f newAabbMax) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
//			unsigned short int quantizedAabbMin[3];
//			unsigned short int quantizedAabbMax[3];
//			quantizeWithClamp(quantizedAabbMin,newAabbMin);
//			quantizeWithClamp(quantizedAabbMax,newAabbMax);
//			for (int i=0;i<3;i++)
//			{
//				if (m_quantizedContiguousNodes[nodeIndex].m_quantizedAabbMin[i] > quantizedAabbMin[i])
//					m_quantizedContiguousNodes[nodeIndex].m_quantizedAabbMin[i] = quantizedAabbMin[i];
//
//				if (m_quantizedContiguousNodes[nodeIndex].m_quantizedAabbMax[i] < quantizedAabbMax[i])
//					m_quantizedContiguousNodes[nodeIndex].m_quantizedAabbMax[i] = quantizedAabbMax[i];
//
//			}
		}
		else {
			//non-quantized
			VectorUtil.setMin(contiguousNodes.get(nodeIndex).aabbMinOrg, newAabbMin);
			VectorUtil.setMax(contiguousNodes.get(nodeIndex).aabbMaxOrg, newAabbMax);
		}
	}
	
	public void swapLeafNodes(int i, int splitIndex) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//btQuantizedBvhNode tmp = m_quantizedLeafNodes[i];
			//m_quantizedLeafNodes[i] = m_quantizedLeafNodes[splitIndex];
			//m_quantizedLeafNodes[splitIndex] = tmp;
		}
		else {
			// JAVA NOTE: changing reference instead of copy
			OptimizedBvhNode tmp = leafNodes.get(i);
			leafNodes.set(i, leafNodes.get(splitIndex));
			leafNodes.set(splitIndex, tmp);
		}
	}

	public void assignInternalNodeFromLeafNode(int internalNode, int leafNodeIndex) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
			//m_quantizedContiguousNodes[internalNode] = m_quantizedLeafNodes[leafNodeIndex];
		}
		else {
			contiguousNodes.get(internalNode).set(leafNodes.get(leafNodeIndex));
		}
	}

	private static class NodeTriangleCallback implements InternalTriangleIndexCallback {
		public List<OptimizedBvhNode> triangleNodes;
		
		public NodeTriangleCallback(List<OptimizedBvhNode> triangleNodes) {
			this.triangleNodes = triangleNodes;
		}

		private Vector3f aabbMin = new Vector3f(), aabbMax = new Vector3f();
		
		public void internalProcessTriangleIndex(Vector3f[] triangle, int partId, int triangleIndex) {
			OptimizedBvhNode node = new OptimizedBvhNode();
			aabbMin.set(1e30f, 1e30f, 1e30f);
			aabbMax.set(-1e30f, -1e30f, -1e30f);
			VectorUtil.setMin(aabbMin, triangle[0]);
			VectorUtil.setMax(aabbMax, triangle[0]);
			VectorUtil.setMin(aabbMin, triangle[1]);
			VectorUtil.setMax(aabbMax, triangle[1]);
			VectorUtil.setMin(aabbMin, triangle[2]);
			VectorUtil.setMax(aabbMax, triangle[2]);

			// with quantization?
			node.aabbMinOrg.set(aabbMin);
			node.aabbMaxOrg.set(aabbMax);

			node.escapeIndex = -1;

			// for child nodes
			node.subPart = partId;
			node.triangleIndex = triangleIndex;
			triangleNodes.add(node);
		}
	}
	
	public void build(StridingMeshInterface triangles, boolean useQuantizedAabbCompression, Vector3f bvhAabbMin, Vector3f bvhAabbMax) {
		stack.vectors.push();
		try {
			this.useQuantization = useQuantizedAabbCompression;

			// NodeArray	triangleNodes;

			int numLeafNodes = 0;

			if (useQuantization) {
				throw new UnsupportedOperationException();
	//			// initialize quantization values
	//			setQuantizationValues(bvhAabbMin,bvhAabbMax);
	//
	//			QuantizedNodeTriangleCallback	callback(m_quantizedLeafNodes,this);
	//
	//
	//			triangles->InternalProcessAllTriangles(&callback,m_bvhAabbMin,m_bvhAabbMax);
	//
	//			//now we have an array of leafnodes in m_leafNodes
	//			numLeafNodes = m_quantizedLeafNodes.size();
	//
	//
	//			m_quantizedContiguousNodes.resize(2*numLeafNodes);
			}
			else {
				NodeTriangleCallback callback = new NodeTriangleCallback(leafNodes);

				Vector3f aabbMin = stack.vectors.get(-1e30f, -1e30f, -1e30f);
				Vector3f aabbMax = stack.vectors.get(1e30f, 1e30f, 1e30f);

				triangles.internalProcessAllTriangles(callback, aabbMin, aabbMax);

				// now we have an array of leafnodes in m_leafNodes
				numLeafNodes = leafNodes.size();

				// TODO: check
				//contiguousNodes.resize(2*numLeafNodes);
				MiscUtil.resize(contiguousNodes, 2 * numLeafNodes, OptimizedBvhNode.class);
			}

			curNodeIndex = 0;

			buildTree(0, numLeafNodes);

	//		//  if the entire tree is small then subtree size, we need to create a header info for the tree
	//		if(useQuantization && SubtreeHeaders.size() == 0)
	//		{
	//			btBvhSubtreeInfo& subtree = m_SubtreeHeaders.expand();
	//			subtree.setAabbFromQuantizeNode(m_quantizedContiguousNodes[0]);
	//			subtree.m_rootNodeIndex = 0;
	//			subtree.m_subtreeSize = m_quantizedContiguousNodes[0].isLeafNode() ? 1 : m_quantizedContiguousNodes[0].getEscapeIndex();
	//		}

			// PCK: update the copy of the size
			subtreeHeaderCount = SubtreeHeaders.size();

			// PCK: clear m_quantizedLeafNodes and m_leafNodes, they are temporary
	//		quantizedLeafNodes.clear();
			leafNodes.clear();
		}
		finally {
			stack.vectors.pop();
		}
	}
	
	public void refit(StridingMeshInterface meshInterface) {
		if (useQuantization) {
			throw new UnsupportedOperationException();
//			//calculate new aabb
//			btVector3 aabbMin,aabbMax;
//			meshInterface->calculateAabbBruteForce(aabbMin,aabbMax);
//
//			setQuantizationValues(aabbMin,aabbMax);
//
//			updateBvhNodes(meshInterface,0,m_curNodeIndex,0);
//
//			///now update all subtree headers
//
//			int i;
//			for (i=0;i<m_SubtreeHeaders.size();i++)
//			{
//				btBvhSubtreeInfo& subtree = m_SubtreeHeaders[i];
//				subtree.setAabbFromQuantizeNode(m_quantizedContiguousNodes[subtree.m_rootNodeIndex]);
//			}
//
		}
		else {

		}
	}
	
	protected void buildTree(int startIndex, int endIndex) {
		stack.vectors.push();
		try {
			//#ifdef DEBUG_TREE_BUILDING
			if (DEBUG_TREE_BUILDING) {
				gStackDepth++;
				if (gStackDepth > gMaxStackDepth) {
					gMaxStackDepth = gStackDepth;
				}
			}
			//#endif //DEBUG_TREE_BUILDING

			int splitAxis, splitIndex, i;
			int numIndices = endIndex - startIndex;
			int curIndex = curNodeIndex;

			assert (numIndices > 0);

			if (numIndices == 1) {
				//#ifdef DEBUG_TREE_BUILDING
				if (DEBUG_TREE_BUILDING) {
					gStackDepth--;
				}
				//#endif //DEBUG_TREE_BUILDING

				assignInternalNodeFromLeafNode(curNodeIndex, startIndex);

				curNodeIndex++;
				return;
			}
			// calculate Best Splitting Axis and where to split it. Sort the incoming 'leafNodes' array within range 'startIndex/endIndex'.

			splitAxis = calcSplittingAxis(startIndex, endIndex);

			splitIndex = sortAndCalcSplittingIndex(startIndex, endIndex, splitAxis);

			int internalNodeIndex = curNodeIndex;

			setInternalNodeAabbMax(curNodeIndex, stack.vectors.get(-1e30f, -1e30f, -1e30f));
			setInternalNodeAabbMin(curNodeIndex, stack.vectors.get(1e30f, 1e30f, 1e30f));

			for (i = startIndex; i < endIndex; i++) {
				mergeInternalNodeAabb(curNodeIndex, getAabbMin(i), getAabbMax(i));
			}

			curNodeIndex++;

			//internalNode->m_escapeIndex;

			int leftChildNodexIndex = curNodeIndex;

			//build left child tree
			buildTree(startIndex, splitIndex);

			int rightChildNodexIndex = curNodeIndex;
			// build right child tree
			buildTree(splitIndex, endIndex);

			//#ifdef DEBUG_TREE_BUILDING
			if (DEBUG_TREE_BUILDING) {
				gStackDepth--;
			}
			//#endif //DEBUG_TREE_BUILDING

			int escapeIndex = curNodeIndex - curIndex;

			if (useQuantization) {
				throw new UnsupportedOperationException();
	//			//escapeIndex is the number of nodes of this subtree
	//			const int sizeQuantizedNode =sizeof(btQuantizedBvhNode);
	//			const int treeSizeInBytes = escapeIndex * sizeQuantizedNode;
	//			if (treeSizeInBytes > MAX_SUBTREE_SIZE_IN_BYTES)
	//			{
	//				updateSubtreeHeaders(leftChildNodexIndex,rightChildNodexIndex);
	//			}
			}

			setInternalNodeEscapeIndex(internalNodeIndex, escapeIndex);
		}
		finally {
			stack.vectors.pop();
		}
	}

	protected int sortAndCalcSplittingIndex(int startIndex, int endIndex, int splitAxis) {
		stack.vectors.push();
		try {
			int i;
			int splitIndex = startIndex;
			int numIndices = endIndex - startIndex;
			float splitValue;

			Vector3f means = stack.vectors.get(0f, 0f, 0f);
			Vector3f center = stack.vectors.get();
			for (i = startIndex; i < endIndex; i++) {
				center.add(getAabbMax(i), getAabbMin(i));
				center.scale(0.5f);
				means.add(center);
			}
			means.scale(1f / (float) numIndices);

			splitValue = VectorUtil.getCoord(means, splitAxis);

			//sort leafNodes so all values larger then splitValue comes first, and smaller values start from 'splitIndex'.
			for (i = startIndex; i < endIndex; i++) {
				//Vector3f center = new Vector3f();
				center.add(getAabbMax(i), getAabbMin(i));
				center.scale(0.5f);

				if (VectorUtil.getCoord(center, splitAxis) > splitValue) {
					// swap
					swapLeafNodes(i, splitIndex);
					splitIndex++;
				}
			}

			// if the splitIndex causes unbalanced trees, fix this by using the center in between startIndex and endIndex
			// otherwise the tree-building might fail due to stack-overflows in certain cases.
			// unbalanced1 is unsafe: it can cause stack overflows
			// bool unbalanced1 = ((splitIndex==startIndex) || (splitIndex == (endIndex-1)));

			// unbalanced2 should work too: always use center (perfect balanced trees)	
			// bool unbalanced2 = true;

			// this should be safe too:
			int rangeBalancedIndices = numIndices / 3;
			boolean unbalanced = ((splitIndex <= (startIndex + rangeBalancedIndices)) || (splitIndex >= (endIndex - 1 - rangeBalancedIndices)));

			if (unbalanced) {
				splitIndex = startIndex + (numIndices >> 1);
			}

			boolean unbal = (splitIndex == startIndex) || (splitIndex == (endIndex));
			assert (!unbal);

			return splitIndex;
		}
		finally {
			stack.vectors.pop();
		}
	}

	protected int calcSplittingAxis(int startIndex, int endIndex) {
		stack.vectors.push();
		try {
			int i;

			Vector3f means = stack.vectors.get(0f, 0f, 0f);
			Vector3f variance = stack.vectors.get(0f, 0f, 0f);
			int numIndices = endIndex - startIndex;

			Vector3f center = stack.vectors.get();
			for (i = startIndex; i < endIndex; i++) {
				center.add(getAabbMax(i), getAabbMin(i));
				center.scale(0.5f);
				means.add(center);
			}
			means.scale(1f / (float) numIndices);

			Vector3f diff2 = stack.vectors.get();
			for (i = startIndex; i < endIndex; i++) {
				center.add(getAabbMax(i), getAabbMin(i));
				center.scale(0.5f);
				diff2.sub(center, means);
				//diff2 = diff2 * diff2;
				VectorUtil.mul(diff2, diff2, diff2);
				variance.add(diff2);
			}
			variance.scale(1f / ((float) numIndices - 1));

			return VectorUtil.maxAxis(variance);
		}
		finally {
			stack.vectors.pop();
		}
	}

	public void reportAabbOverlappingNodex(NodeOverlapCallback nodeCallback, Vector3f aabbMin, Vector3f aabbMax) {
		// either choose recursive traversal (walkTree) or stackless (walkStacklessTree)

		if (useQuantization) {
			throw new UnsupportedOperationException();
//			///quantize query AABB
//			unsigned short int quantizedQueryAabbMin[3];
//			unsigned short int quantizedQueryAabbMax[3];
//			quantizeWithClamp(quantizedQueryAabbMin,aabbMin);
//			quantizeWithClamp(quantizedQueryAabbMax,aabbMax);
//
//			switch (m_traversalMode)
//			{
//			case TRAVERSAL_STACKLESS:
//					walkStacklessQuantizedTree(nodeCallback,quantizedQueryAabbMin,quantizedQueryAabbMax,0,m_curNodeIndex);
//				break;
//			case TRAVERSAL_STACKLESS_CACHE_FRIENDLY:
//					walkStacklessQuantizedTreeCacheFriendly(nodeCallback,quantizedQueryAabbMin,quantizedQueryAabbMax);
//				break;
//			case TRAVERSAL_RECURSIVE:
//				{
//					const btQuantizedBvhNode* rootNode = &m_quantizedContiguousNodes[0];
//					walkRecursiveQuantizedTreeAgainstQueryAabb(rootNode,nodeCallback,quantizedQueryAabbMin,quantizedQueryAabbMax);
//				}
//				break;
//			default:
//				//unsupported
//				btAssert(0);
//			}
		}
		else {
			walkStacklessTree(nodeCallback, aabbMin, aabbMax);
		}
	}
	
	protected void walkStacklessTree(NodeOverlapCallback nodeCallback, Vector3f aabbMin, Vector3f aabbMax) {
		assert (!useQuantization);

		// JAVA NOTE: rewritten
		OptimizedBvhNode rootNode = null;//contiguousNodes.get(0);
		int rootNode_index = 0;

		int escapeIndex, curIndex = 0;
		int walkIterations = 0;
		boolean isLeafNode;
		//PCK: unsigned instead of bool
		//unsigned aabbOverlap;
		boolean aabbOverlap;

		while (curIndex < curNodeIndex) {
			// catch bugs in tree data
			assert (walkIterations < curNodeIndex);

			walkIterations++;

			rootNode = contiguousNodes.get(rootNode_index);

			aabbOverlap = AabbUtil2.testAabbAgainstAabb2(aabbMin, aabbMax, rootNode.aabbMinOrg, rootNode.aabbMaxOrg);
			isLeafNode = (rootNode.escapeIndex == -1);

			// PCK: unsigned instead of bool
			if (isLeafNode && (aabbOverlap/* != 0*/)) {
				nodeCallback.processNode(rootNode.subPart, rootNode.triangleIndex);
			}

			rootNode = null;

			//PCK: unsigned instead of bool
			if ((aabbOverlap/* != 0*/) || isLeafNode) {
				rootNode_index++;
				curIndex++;
			}
			else {
				escapeIndex = /*rootNode*/ contiguousNodes.get(rootNode_index).escapeIndex;
				rootNode_index += escapeIndex;
				curIndex += escapeIndex;
			}
		}
		if (maxIterations < walkIterations) {
			maxIterations = walkIterations;
		}
	}

	public void reportRayOverlappingNodex(NodeOverlapCallback nodeCallback, Vector3f raySource, Vector3f rayTarget) {
		stack.vectors.push();
		try {
			boolean fast_path = useQuantization && traversalMode == TraversalMode.TRAVERSAL_STACKLESS;
			if (fast_path) {
				throw new UnsupportedOperationException();
				//walkStacklessQuantizedTreeAgainstRay(nodeCallback, raySource, rayTarget, btVector3(0, 0, 0), btVector3(0, 0, 0), 0, m_curNodeIndex);
			}
			else {
				/* Otherwise fallback to AABB overlap test */
				Vector3f aabbMin = stack.vectors.get(raySource);
				Vector3f aabbMax = stack.vectors.get(raySource);
				VectorUtil.setMin(aabbMin, rayTarget);
				VectorUtil.setMax(aabbMax, rayTarget);
				reportAabbOverlappingNodex(nodeCallback, aabbMin, aabbMax);
			}
		}
		finally {
			stack.vectors.pop();
		}
	}

	public void reportBoxCastOverlappingNodex(NodeOverlapCallback nodeCallback, Vector3f raySource, Vector3f rayTarget, Vector3f aabbMin, Vector3f aabbMax) {
		stack.vectors.push();
		try {
			boolean fast_path = useQuantization && traversalMode == TraversalMode.TRAVERSAL_STACKLESS;
			if (fast_path) {
				throw new UnsupportedOperationException();
				//walkStacklessQuantizedTreeAgainstRay(nodeCallback, raySource, rayTarget, aabbMin, aabbMax, 0, m_curNodeIndex);
			}
			else {
				/* Slow path:
				Construct the bounding box for the entire box cast and send that down the tree */
				Vector3f qaabbMin = stack.vectors.get(raySource);
				Vector3f qaabbMax = stack.vectors.get(raySource);
				VectorUtil.setMin(qaabbMin, rayTarget);
				VectorUtil.setMax(qaabbMax, rayTarget);
				qaabbMin.add(aabbMin);
				qaabbMax.add(aabbMax);
				reportAabbOverlappingNodex(nodeCallback, qaabbMin, qaabbMax);
			}
		}
		finally {
			stack.vectors.pop();
		}
	}
	
}
